import logging
import polars as pl

logger = logging.getLogger(__name__)

def validate(df: pl.DataFrame, options):
    # Validate that all feature columns are present
    for fcol in options['input']['columns']['features']:
        if fcol not in df.columns:
            logger.warning(f"Feature column `{fcol}` not in input DataFrame.")

    # Validate that at least score or fdr or training_flag column is present
    col_score = options['input']['columns']['score']
    if col_score not in df.columns:
        raise XiRescoreInputError(f"Score column `{col_score}` was not found.")

    # Validate that sequence columns are present
    col_seq1 = options['input']['columns']['base_sequence_p1']
    col_seq2 = options['input']['columns']['base_sequence_p2']
    if col_seq1 not in df.columns:
        raise XiRescoreInputError(f"Missing sequence column `{col_seq1}`.")
    if col_seq2 not in df.columns:
        raise XiRescoreInputError(f"Missing sequence column `{col_seq2}`.")

    # Validate that target column is present
    col_target = options['input']['columns']['is_tt']
    if col_target not in df.columns:
        raise XiRescoreInputError(f"Missing target column `{col_target}`.")


class XiRescoreInputError(Exception):
    pass



import polars as pl
from xifdr import fdr

def generate(df: pl.DataFrame, options: dict, do_self_between=False, do_fdr=False) -> pl.DataFrame:
    input_cols = options['input']['columns']
    consts = options['input']['constants']
    cols_spectra = input_cols['spectrum_id']
    col_score = input_cols['score']
    # Generate base sequences
    if input_cols['base_sequence_p1'] not in df.columns:
        # Remove modifications
        df = df.with_columns(
            pl.col(input_cols['sequence_p1'])\
                .str.replace_all('\\(.*\\)', '')\
                .str.replace_all('\\[.*\\]', '')\
                .str.replace_all('[^A-Z]', '')\
                .alias(input_cols['base_sequence_p1'])
        )
    if input_cols['base_sequence_p2'] not in df.columns:
        # Remove modifications
        df = df.with_columns(
            pl.col(input_cols['sequence_p2'])\
                .str.replace_all('\\(.*\\)', '')\
                .str.replace_all('\\[.*\\]', '')\
                .str.replace_all('[^A-Z]', '')\
                .alias(input_cols['base_sequence_p2'])
        )
    # Generate top_ranking
    if input_cols['top_ranking'] not in df.columns:
        df_max = df.group_by(cols_spectra).agg(
            pl.col(col_score).max().alias(f'{col_score}_max')
        )
        df = df.join(
            df_max,
            on=list(cols_spectra)
        )
        df = df.with_columns(
            top_ranking=pl.col(col_score) == pl.col(f'{col_score}_max')
        )
    # Generate decoy_p1/2
    ordered_decoy_class = (consts['td_class'] is not None) and (consts['dt_class'] is not None)
    if input_cols['decoy_class'] in df.columns and ordered_decoy_class:
        df = df.with_columns(
            (
                pl.col(input_cols['decoy_class']).is_in([consts['tt_class'], consts['td_class']])
            ).alias(input_cols['decoy_p1']),
            (
                pl.col(input_cols['decoy_class']).is_in([consts['dd_class'], consts['dt_class']])
            ).alias(input_cols['decoy_p2']),
        )
    # Generate decoy_class column from decoy_p1 and decoy_p2
    if input_cols['decoy_class'] not in df.columns:
        tt_expr = pl.col(input_cols['decoy_p1']).not_() & pl.col(input_cols['decoy_p2']).not_()
        dd_expr = pl.col(input_cols['decoy_p1']) & pl.col(input_cols['decoy_p2'])
        df = df.with_columns(
            pl.when(tt_expr).then(pl.lit(consts['tt_class']))\
                .when(dd_expr).then(pl.lit(consts['dd_class']))\
                .otherwise(pl.lit(consts['td_class']))\
                .alias(input_cols['decoy_class'])
        )
    # Generate target column from decoy_class
    if input_cols['is_tt'] not in df.columns:
        df = df.with_columns(
            (pl.col(input_cols['decoy_class']) == consts['tt_class']).alias(input_cols['is_tt'])
        )
    # Calculte self_between from protein_p1, and protein_p2
    if do_self_between and input_cols['self_between'] not in df.columns:
        protein_p1_list = pl.col(input_cols['protein_p1'])
        protein_p2_list = pl.col(input_cols['protein_p2'])
        if type(df[input_cols['protein_p1']].dtype) is pl.String:
            protein_p1_list = protein_p1_list.str.split(';')
        if type(df[input_cols['protein_p2']].dtype) is pl.String:
            protein_p2_list = protein_p2_list.str.split(';')
        protein_p1_list = protein_p1_list.list.eval(
            pl.element().str.replace_all(options['input']['constants']['decoy_adjunct'], '')
        )
        protein_p2_list = protein_p2_list.list.eval(
            pl.element().str.replace_all(options['input']['constants']['decoy_adjunct'], '')
        )
        overlap_expr = protein_p1_list.list.set_intersection(protein_p2_list)
        df = df.with_columns(
            fdr_group = pl.when(
                overlap_expr.list.len() == 0
            ).then(
                pl.lit('between')
            ).otherwise(
                pl.lit('self')
            )
        )
    # Calculate fdr from self_between and score
    if do_fdr and input_cols['fdr'] not in df.columns:
        fdr_ser = fdr.single_grouped_fdr(
            df.with_columns(
                score=pl.col(input_cols['score']),
                decoy_class=pl.col(input_cols['decoy_class']).replace(
                    [consts['tt_class'], consts['td_class'], consts['dt_class'], consts['dd_class']],
                    ['TT', 'TD', 'DT', 'DD']
                ),
                fdr_group=pl.col(input_cols['self_between']),
            ).with_columns(
                TT=pl.col('decoy_class')=='TT',
                TD=pl.col('decoy_class')=='TD',
                DD=pl.col('decoy_class')=='DD',
            ),
        )
        df = df.with_columns(
            fdr=fdr_ser
        )
    return df



import multiprocess as mp
from math import ceil
import logging

import numpy as np
import polars as pl
import scipy

from xirescore import async_result_resolving


logger = logging.getLogger(__name__)

def rescore(models,
            df,
            rescore_col,
            apply_logit=False,
            max_cpu=-1):
    if not isinstance(df, pl.DataFrame):
        df: pl.DataFrame = pl.DataFrame(df)
    n_procs = max_cpu
    if n_procs < 1:
        n_procs = int(mp.cpu_count() - 1)
    n_models = len(models)
    n_dataslices = ceil(n_procs/n_models)
    slice_size = ceil(len(df) / n_dataslices)
    logger.debug(f'Split {len(df):,.0f} samples in {n_dataslices} slices of {slice_size}')

    # Slice input data for multiprocessing
    dataslices = [
        df[
            i*slice_size:(i+1)*slice_size
        ]
        for i in range(n_dataslices)
    ]

    # Apply each classifier to each data slice
    with mp.get_context("spawn").Pool(n_procs) as pool:
        async_results = []
        for slice in dataslices:
            for clf in models:
                if "Perceptron" in str(clf):
                    async_results.append(
                        pool.apply_async(
                            clf.decision_function,
                            (slice,),
                        )
                    )
                elif "tensorflow" not in str(clf):
                    async_results.append(
                        pool.apply_async(
                            clf.predict_proba,
                            (slice,),
                        )
                    )
                else:
                    async_results.append(
                        pool.apply_async(
                            clf.predict_proba,
                            (slice,),
                        )
                    )

        sync_results = async_result_resolving.resolve(async_results)
        slice_results: list = [
            # Join slices into a single array of all model predictions each
            np.array([
                sync_results[
                    (slice_i*n_models)+model_i
                ][:, 1]  # Take only prediction for class 1 (aka target)
                for model_i in range(n_models)
            ])
            for slice_i in range(n_dataslices)
        ]

    # Apply logit
    if apply_logit:
        for i, sr in enumerate(slice_results):
            slice_results[i] = scipy.special.logit(sr)

    # Init result DF
    np_rescore = np.zeros((len(df), 0))

    # Fill result DF
    for i_m, model in enumerate(models):
        rescores_m = [
            slice_results[i_s][i_m]
            for i_s in range(n_dataslices)
        ]
        np_rescore = np.hstack([
            np_rescore,
            np.concatenate(rescores_m).reshape(len(df), 1)
        ])

    df_rescore = pl.DataFrame(
        np_rescore,
        schema=[(f'{rescore_col}_{i_m}', pl.Float64) for i_m, _ in enumerate(models)],
    )

    # Calculate mean score and standard deviation
    df_rescore = df_rescore.with_columns(
        df_rescore.mean_horizontal().alias(rescore_col),
        df_rescore.transpose().std().transpose().to_series().alias(f'{rescore_col}_std'),
    )

    return df_rescore



import numpy as np

default_options = {
    'input': {
        'columns': {
            # Boolean column indicating if a match has
            # the highest score for the specific spectrum.
            # If you don't allow for multiple matches for one spectrum
            # this column would only contain ``True`` values.
            'top_ranking': 'top_ranking',

            # The main score to use for FDR calculation and therefore training data selection.
            'score': 'match_score',

            # The AA sequence of the first peptide with modifications.
            'sequence_p1': 'sequence_p1',

            # The AA sequence of the second peptide with modifications.
            'sequence_p2': 'sequence_p2',

            # The AA sequence of the first peptide without any modifications.
            # If not found, created by removing anything not A-Z and anything
            # in brackets from sequence_p1.
            'base_sequence_p1': 'base_sequence_p1',

            # The AA sequence of the second peptide without any modifications.
            # If not found, created by removing anything not A-Z and anything
            # in brackets from sequence_p2.
            'base_sequence_p2': 'base_sequence_p2',

            # Semicolon separated list of proteins associated with first peptide.
            'protein_p1': 'protein_p1',

            # Semicolon separated list of proteins associated with second peptide.
            'protein_p2': 'protein_p2',

            # Boolean column indicating if the first peptide is a decoy.
            'decoy_p1': 'is_decoy_p1',

            # Boolean column indicating if the second peptide is a decoy.
            'decoy_p2': 'is_decoy_p2',

            # FDR calculated for the top ranking matches.
            # Will be generated if not present in input.
            'fdr': 'fdr',

            # Information if a CSM is a self (homomeric) or between (heteromeric) match.
            # Specific classification values are defined under input.constants.
            'self_between': 'fdr_group',

            # Boolean columns
            'is_tt': 'isTT',

            # Decoy class information. Indicates if the CSM is a target-target ("TT"),
            # target-decoy ("TD") or decoy-decoy ("DD")
            'decoy_class': 'decoy_class',

            # Precursor charge (used for unique training sample filter)
            'charge': 'charge',

            # Link position in peptide 1 (used for unique training sample filter)
            'link_pos_p1': 'link_pos_p1',

            # Link position in peptide 2 (used for unique training sample filter)
            'link_pos_p2': 'link_pos_p2',

            # Feature columns prefix. All columns starting with this prefix
            # will be considered feature columns.
            'feature_prefix': 'feature_',

            # Additional columns to be considered features
            'features': [],

            # Columns that uniquely identify a CSM. If None all columns of the input will be used.
            'csm_id': None,

            # Columns that uniquely identify a spectrum. Will be used for new top_ranking calculation.
            'spectrum_id': ['spectrum_id'],
        },
        'constants': {
            # Self classification value for column self_between.
            'self': 'self',

            # Between classification value for column self_between.
            'between': 'between',

            # String added to decoy protein names. Will be removed for self/between classification.
            'decoy_adjunct': 'REV_',

            # Decoy class values for decoy_class column.
            'tt_class': 'TT',
            'td_class': 'TD',
            'dt_class': None,  # Optional, only for creating decoy_p1/2 columns from decoy_class.
            'dd_class': 'DD',
        },
    },
    'rescoring': {
        # Imputer from `sklearn.impute` to use
        'imputer': None,

        # Keyword arguments for the imputer
        'imputer_kwargs': {},

        # PCA number of components to keep. If None, perform no PCA.
        'pca_n_components': None,

        # FDR threshold for training samples.
        'train_fdr_threshold': 0.01,

        # Method on how to sample the training data.
        'train_selection_mode': 'self-targets-all-decoys',

        # Number of samples used for training data selection.
        'top_sample_size': 1_000_000,

        # Maximum number of training samples
        'train_size_max': 20_000,

        # Whether to train on unique CSMs
        'train_unique_csms': True,

        # Number of spectra that will be loaded and rescored at once.
        'spectra_batch_size': 100_000,

        # SKLearn module of the desired classifier.
        'model_class': 'linear_model',

        # Class name of the desired classifier.
        'model_name': 'LogisticRegression',

        # Name of the desired metric to evaluate models from sklearn.metrics.
        'metric_name': 'log_loss',

        # Boolean indicating if the metric is better when smaller.
        'minimize_metric': True,

        # Model parameters for hyperparameter optimization.
        'model_params': {
            "C": np.logspace(-3, 2, 6),
            "solver": ["liblinear"],
            "penalty": ["l1", "l2"],
            "class_weight": ["balanced", None, {0: 2, 1: 1}],
            "random_state": [0],
        },

        # Number of k-fold splits.
        'n_splits': 5,

        # Limit of how many CPUs to use.
        'max_jobs': -1,

        # Random seed for Python random module and numpy.
        'random_seed': 0,

        # Boolean indicating if logit should be applied to the rescoring result.
        # This can make the rescore more normally distributed with certain
        # classifiers. (E.g. LinearRegression or SVC)
        'logit_result': True,

        # Scaler to use for the features
        'scaler': 'StandardScaler',

        # Scaler's keyword arguments
        'scaler_params': {}
    },
    'output': {
        'columns': {
            # Name of the column containing the new score.
            # Also base name for each model's separate prediction
            # and training slice information.
            'rescore': 'rescore',
        }
    },
}



import logging
from logging import Logger

import polars as pl


logger = logging.getLogger(__name__)

def get_features(df: pl.DataFrame, options: dict, filter_nan=True):
    features_const = options['input']['columns']['features']
    feat_prefix = options['input']['columns']['feature_prefix']
    features_prefixes = [
        c for c in df.columns if str(c).startswith(feat_prefix)
    ]
    features = list(set(features_const + features_prefixes))

    absent_features = [
        f
        for f in features
        if f not in df.columns
    ]
    nan_features = []
    if filter_nan:
        nan_features = [
            f
            for f in features
            if (f not in absent_features) and any(df[f].is_null())
        ]
    features = [
        f
        for f in features
        if f not in (absent_features + nan_features)
    ]

    if len(nan_features) > 0:
        logger.warning(f"Dropped features with NaN values: {nan_features}")
    if len(absent_features) > 0:
        logger.warning(f"Did not find some features: {absent_features}")

    return features



import logging

import polars as pl
import numpy as np
from sklearn import impute

from xirescore import readers
from xirescore.column_generating import generate as generate_columns
from xirescore.feature_scaling import get_transformers


logger = logging.getLogger(__name__)

def select(input_data, options):
    """
    Select training data for Crosslink MS Machine Learning based on specified options.

    Parameters:
    - input_data: The input data for selection.
    - options: Dictionary containing various configuration options for the selection process.
    - logger: Logger instance for logging information and debugging.

    Returns:
    - A pandas DataFrame containing the selected training data.
    """

    # Extract options
    selection_mode = options['rescoring']['train_selection_mode']
    train_unique = options['rescoring']['train_unique_csms']
    train_size_max = options['rescoring']['train_size_max']
    top_sample_size = options['rescoring']['top_sample_size']
    top_ranking_col = options['input']['columns']['top_ranking']
    col_sequence_p1 = options['input']['columns']['base_sequence_p1']
    col_sequence_p2 = options['input']['columns']['base_sequence_p2']
    col_base_sequence_p2 = options['input']['columns']['base_sequence_p2']
    seed = options['rescoring']['random_seed']
    col_self_between = options['input']['columns']['self_between']
    col_fdr = options['input']['columns']['fdr']
    col_native_score = options['input']['columns']['score']
    col_target = options['input']['columns']['is_tt']
    col_link_pos_p1 = options['input']['columns']['link_pos_p1']
    col_link_pos_p2 = options['input']['columns']['link_pos_p2']
    col_charge = options['input']['columns']['charge']
    fdr_cutoff = options['rescoring']['train_fdr_threshold']
    val_self = options['input']['constants']['self']

    # Generate schema overrides
    float_cols = options['input']['columns']['features']
    float_cols.append(
        options['input']['columns']['score']
    )
    schema_overrides = {
        c: pl.Float64
        for c in float_cols
    }

    # Read input data
    df = readers.read_sample(
        input_data,
        sample=top_sample_size,
        top_ranking_col=top_ranking_col,
        sequence_p2_col=col_base_sequence_p2,
        only_top_ranking=True,
        only_pairs=True,
        schema_overrides=schema_overrides,
    )
    logger.debug(f'Fetched {len(df)} top ranking samples')

    if train_unique:
        unique_cols = [
            col_sequence_p1,
            col_sequence_p2,
            col_link_pos_p1,
            col_link_pos_p2,
            col_charge,
        ]
        df = df.sort(col_native_score, descending=True).unique(subset=unique_cols, keep='first')

    # Generate needed columns
    df = generate_columns(df, options=options, do_fdr=True, do_self_between=True)

    # Get scaler
    imputer, scaler, features = get_transformers(df, options)

    # Selection mode: self-targets-all-decoys
    logger.info(f'Use selection mode {selection_mode}')
    if selection_mode == 'self-targets-all-decoys':
        # Create filters
        filter_self = pl.col(col_self_between) == val_self
        filter_fdr = pl.col(col_fdr) <= fdr_cutoff
        filter_target = pl.col(col_target)

        # Max target size
        target_max = int(train_size_max / 2)

        # Get self targets
        train_self_targets = df.filter(filter_fdr & filter_target & filter_self)
        if len(train_self_targets) > target_max:
            train_self_targets = train_self_targets.sample(target_max, seed=seed)
        logger.info(f'Taking {len(train_self_targets)} self targets below {fdr_cutoff} FDR')

        # Get between targets
        train_between_targets = df.filter(filter_fdr & filter_target & ~filter_self)
        sample_min = min(
            len(train_between_targets),
            int(train_size_max/2)-len(train_self_targets),
        )
        train_between_targets = train_between_targets.sample(sample_min, seed=seed)
        logger.info(f'Taking {len(train_between_targets)} between targets below {fdr_cutoff} FDR')

        # Get self decoy-x
        train_self_decoys = df.filter(filter_self & ~filter_target)
        sample_min = min(
            len(train_self_decoys),
            int(train_size_max/4)
        )
        train_self_decoys = train_self_decoys.sample(sample_min, seed=seed)
        logger.info(f'Taking {len(train_self_decoys)} self decoys.')

        # Get between decoy-x
        train_between_decoys = df.filter((~filter_self) & (~filter_target))
        sample_min = min(
            len(train_between_decoys),
            int(train_size_max/2)-len(train_self_decoys),
        )
        train_between_decoys = train_between_decoys.sample(sample_min, seed=seed)
        logger.info(f'Taking {len(train_between_decoys)} between decoys.')

        train_data_df = pl.concat([
            train_self_targets,
            train_between_targets,
            train_self_decoys,
            train_between_decoys
        ])

    # Selection mode: self-targets-capped-decoys
    elif selection_mode == 'self-targets-capped-decoys':
        # Create filters
        filter_self = pl.col(col_self_between) == val_self
        filter_fdr = pl.col(col_fdr) <= fdr_cutoff
        filter_target = pl.col(col_target)

        # Max target size
        target_max = int(train_size_max / 2)

        # Get self targets
        train_self_targets = df.filter(filter_fdr & filter_target & filter_self)
        if len(train_self_targets) > target_max:
            train_self_targets = train_self_targets.sample(target_max, seed=seed)
        logger.info(f'Taking {len(train_self_targets)} self targets below {fdr_cutoff} FDR')

        # Get between targets
        train_between_targets = df.filter(filter_fdr & filter_target & ~filter_self)
        sample_min = min(
            len(train_between_targets),
            int(train_size_max/2)-len(train_self_targets),
        )
        train_between_targets = train_between_targets.sample(sample_min, seed=seed)
        logger.info(f'Taking {len(train_between_targets)} between targets below {fdr_cutoff} FDR')

        # Get capped decoy-x
        all_target = df.filter(filter_target)
        all_decoy = df.filter(~filter_target)
        _, hist_bins = np.histogram(df[col_native_score].to_numpy(), bins=1_000)
        hist_tt, _ = np.histogram(all_target[col_native_score].to_numpy(), bins=hist_bins)
        hist_dx, _ = np.histogram(all_decoy[col_native_score].to_numpy(), bins=hist_bins)
        hist_dx_capped = np.minimum(hist_dx, hist_tt)

        # Number of Dx to aim for
        n_decoy = min([
            hist_dx_capped.sum(),
            train_size_max/2
        ])

        # Scale decoy-x histogram
        dx_scale_fact = min(
            1,
            n_decoy / hist_dx_capped.sum(),
        )
        hist_dx_scaled = (hist_dx_capped * dx_scale_fact).round().astype(int)

        train_decoys = pl.DataFrame(schema=df.schema)
        for i, n in enumerate(hist_dx_scaled):
            if n == 0:
                continue
            score_min = hist_bins[i]
            score_max = hist_bins[i + 1]
            bins_samples = all_decoy.filter(
                (pl.col(col_native_score) >= score_min) & (pl.col(col_native_score) < score_max)
            )
            train_decoys = pl.concat(
                [
                    train_decoys,
                    bins_samples.sample(n=n, seed=seed)
                ]
            )

        logger.info(f'Taking {len(train_decoys)} decoys.')

        train_data_df = pl.concat([
            train_self_targets,
            train_between_targets,
            train_decoys]
        )
    else:
        raise TrainDataError(f"Unknown train data selection mode: {selection_mode}.")

    return train_data_df, imputer, scaler, features


class TrainDataError(Exception):
    """Custom exception for train data selection errors."""
    pass



"""
Readers for data inputs
"""
import random
from math import ceil
from typing import Union
from collections.abc import Sequence

import pandas as pd
import polars as pl
from fastparquet import ParquetFile as FPParquetFile


def read_spectra_ids(path, spectra_cols=None, schema_overrides={}) -> list[int]:
    if type(path) in [pd.DataFrame, pl.DataFrame]:
        df = pl.DataFrame(path)
        return df.select(pl.struct(spectra_cols).hash())\
            .unique() \
            .to_series().to_list()

    file_type = get_source_type(path)

    if spectra_cols is None:
        raise ValueError('Filetype {file_type} requires parameter `spectra_cols`!')

    if file_type == 'csv':
        return pl.scan_csv(
                path,
                ignore_errors=True,
                null_values=['∞', '-∞'],
                schema_overrides=schema_overrides,
            ).select(pl.struct(spectra_cols).hash()) \
            .unique() \
            .collect().to_series().to_list()
    if file_type == 'tsv':
        return pl.scan_csv(
                path,
                separator='\t',
                ignore_errors=True,
                null_values=['∞', '-∞'],
                schema_overrides=schema_overrides,
            ).select(pl.struct(spectra_cols).hash()) \
            .unique() \
            .collect().to_series().to_list()
    if file_type == 'parquet':
        return pl.scan_parquet(path) \
            .select(pl.struct(spectra_cols).hash()) \
            .unique() \
            .collect().to_series().to_list()


def read_spectra_range(input: Union[str, pd.DataFrame],
                       spectra_from: int,
                       spectra_to: int,
                       spectra_cols: Sequence = None,
                       sequence_p2_col='sequence_p2',
                       only_pairs=True,
                       schema_overrides={}):
    # Convert pandas to polars
    if type(input) is pd.DataFrame:
        input = pl.DataFrame(input)
    # Handle input DF
    if type(input) is pl.DataFrame:
        filters = (
            (pl.struct(spectra_cols).hash() >= spectra_from) &
            (pl.struct(spectra_cols).hash() <= spectra_to)
        )
        if only_pairs:
            filters &= pl.col(sequence_p2_col).is_not_null()
            filters &= pl.col(sequence_p2_col) != ''
        return input.filter(filters)
    # Handle input path
    file_type = get_source_type(input)
    if file_type == 'csv':
        return read_spectra_range_csv(
            input,
            spectra_from,
            spectra_to,
            spectra_cols=spectra_cols,
            sequence_p2_col=sequence_p2_col,
            only_pairs=only_pairs,
            schema_overrides=schema_overrides,
        )
    if file_type == 'tsv':
        return read_spectra_range_csv(
            input,
            spectra_from,
            spectra_to,
            sep='\t',
            spectra_cols=spectra_cols,
            sequence_p2_col=sequence_p2_col,
            only_pairs=only_pairs,
            schema_overrides=schema_overrides,
        )
    if file_type == 'parquet':
        return read_spectra_range_parquet(
            input,
            spectra_from,
            spectra_to,
            spectra_cols=spectra_cols,
            sequence_p2_col=sequence_p2_col,
            only_pairs=only_pairs,
        )


def read_spectra_range_parquet(path,
                               spectra_from,
                               spectra_to,
                               spectra_cols: Sequence,
                               sequence_p2_col='sequence_p2',
                               only_pairs=True):
    # Filters for spectrum columns
    df = pl.scan_parquet(path)
    # Generate filters
    filters = (
        (pl.struct(spectra_cols).hash() >= spectra_from) &
        (pl.struct(spectra_cols).hash() <= spectra_to)
    )
    # Filter out linear matches
    if only_pairs:
        filters &= pl.col(sequence_p2_col).is_not_null()
        filters &= pl.col(sequence_p2_col) != ''
    return df.filter(filters).collect()


def read_spectra_range_csv(path,
                           spectra_from,
                           spectra_to,
                           spectra_cols: Sequence,
                           sequence_p2_col='sequence_p2',
                           only_pairs=True,
                           sep=',',
                           schema_overrides={}):
    # Filters for spectrum columns
    df = pl.scan_csv(
        path,
        separator=sep,
        ignore_errors=True,
        null_values=['∞', '-∞'],
        schema_overrides=schema_overrides,
    )
    # Generate filters
    filters = (
        (pl.struct(spectra_cols).hash() >= spectra_from) &
        (pl.struct(spectra_cols).hash() <= spectra_to)
    )
    # Filter out linear matches
    if only_pairs:
        filters &= pl.col(sequence_p2_col).is_not_null()
        filters &= pl.col(sequence_p2_col) != ''
    return df.filter(filters).collect()


def get_source_type(path: str):
    if path.lower().endswith('.parquet') or path.lower().endswith('.parquet/'):
        return 'parquet'
    if path.lower().endswith('.tsv') or path.lower().endswith('.tab'):
        return 'tsv'
    if path.lower().endswith('.csv'):
        return 'csv'
    if len(path.split('.')) > 2:
        ext2 = path.split('.')[-2].lower()
        if ext2 == 'csv':
            return 'csv'
        if ext2 == 'tab' or ext2 == 'tsv':
            return 'tsv'
    raise ValueError(f'Unknown file type of {path}')


def read_sample(input_data,
                sample=1_000_000,
                top_ranking_col='top_ranking',
                sequence_p2_col='sequence_p2',
                only_top_ranking=False,
                only_pairs=True,
                schema_overrides={}) -> pl.DataFrame:
    if type(input_data) is pd.DataFrame:
        input_data = pl.DataFrame(input_data)
    if type(input_data) is pl.DataFrame:
        scan_filter = pl.lit(True)
        if only_top_ranking and top_ranking_col in input_data.collect_schema().names():
            scan_filter &= pl.col(top_ranking_col)
        if only_pairs:
            scan_filter &= pl.col(sequence_p2_col).is_not_null()
            scan_filter &= pl.col(sequence_p2_col) != ''
        sample_min = min(
            len(input_data.filter(scan_filter)),
            sample
        )
        return input_data.filter(scan_filter).sample(sample_min)
    file_type = get_source_type(input_data)
    if file_type == 'csv':
        return read_sample_csv(
            input_data,
            sample=sample,
            top_ranking_col=top_ranking_col,
            sequence_p2_col=sequence_p2_col,
            only_top_ranking=only_top_ranking,
            only_pairs=only_pairs,
            schema_overrides=schema_overrides,
        )
    if file_type == 'tsv':
        return read_sample_csv(
            input_data,
            sep='\t',
            sample=sample,
            top_ranking_col=top_ranking_col,
            only_top_ranking=only_top_ranking,
            sequence_p2_col=sequence_p2_col,
            only_pairs=only_pairs,
            schema_overrides=schema_overrides,
        )
    if file_type == 'parquet':
        return read_sample_parquet(
            input_data,
            sample=sample,
            sequence_p2_col=sequence_p2_col,
            only_top_ranking=only_top_ranking,
            top_ranking_col=top_ranking_col,
            only_pairs=only_pairs,
        )


def read_sample_parquet(path: str,
                        sample: int,
                        top_ranking_col='top_ranking',
                        sequence_p2_col='sequence_p2',
                        batch_size=500_000,
                        only_top_ranking=False,
                        only_pairs=True,
                        random_state=random.randint(0, 2**32-1)):
    df_scan = pl.scan_parquet(path)

    scan_filter = pl.lit(True)
    if only_top_ranking and top_ranking_col in df_scan.columns:
        scan_filter &= pl.col(top_ranking_col)
    if only_pairs:
        scan_filter &= pl.col(sequence_p2_col).is_not_null()
        scan_filter &= pl.col(sequence_p2_col) != ''

    df_scan = df_scan.filter(scan_filter)

    n_total = df_scan.select(pl.count()).collect().item()
    every_nth = ceil(n_total / sample)

    return df_scan.gather_every(every_nth).collect()


def read_sample_csv(path,
                    sample,
                    sep=',',
                    top_ranking_col='top_ranking',
                    sequence_p2_col='sequence_p2',
                    only_pairs=True,
                    only_top_ranking=False,
                    schema_overrides={}):
    df_scan = pl.scan_csv(
        path,
        separator=sep,
        ignore_errors=True,
        null_values=['∞', '-∞'],
        schema_overrides=schema_overrides
    )

    scan_filter = pl.lit(True)
    if only_top_ranking and top_ranking_col in df_scan.collect_schema().names():
        scan_filter &= pl.col(top_ranking_col)
    if only_pairs:
        scan_filter &= pl.col(sequence_p2_col).is_not_null()
        scan_filter &= pl.col(sequence_p2_col) != ''

    df_scan = df_scan.filter(scan_filter)
    n_total = df_scan.select(pl.count()).collect().item()
    every_nth = ceil(n_total / sample)

    return df_scan.gather_every(every_nth).collect()



import logging
import functools
from collections.abc import Iterable
import multiprocess as mp
import time
import logging

from tqdm import tqdm


logger = logging.getLogger(__name__)

def resolve(async_results, show_pbar=True, pbar_args=dict()):
    """
    Resolve asynchronous results from multiprocessing with progress bar.
    For compatibility this function returns anything that is not an asynchronous
    result or and array of these.

    :param async_results: Single async result, list of async results or literals.
    :type async_results: Any

    :returns: Resolved result(s) or unprocessed ``async_results`` if it's a literal.
    :rtype: Any
    """
    global logger

    logger.debug("Resolving results")
    if not isinstance(async_results, Iterable):
        logger.debug("Single result")
        if type(async_results) is functools.partial:
            logger.debug("Partial result")
            return async_results()
        if type(async_results) is not mp.pool.AsyncResult:
            logger.debug("Async result")
            return async_results
        return async_results.get()

    logger.debug("List of results")
    pbar = None
    if show_pbar:
        pbar = tqdm(total=len(async_results), dynamic_ncols=True, **pbar_args)
    if type(async_results[0]) is functools.partial:
        logger.debug("Partial results")
        results = []
        for p in async_results:
            results.append(p())
            if pbar:
                pbar.update(1)
        if pbar:
            pbar.close()
        return results

    if type(async_results[0]) is not mp.pool.AsyncResult:
        logger.debug("Static results")
        if pbar:
            pbar.close()
        return async_results

    finished = 0
    results = [None for _ in async_results]
    logger.debug("Async results")
    while finished < len(async_results):
        n_finished = 0
        for i, ar in enumerate(async_results):
            if ar.ready():
                n_finished += 1
                if results[i] is None:
                    results[i] = ar.get()

        new_finished = n_finished - finished
        finished = n_finished

        if pbar:
            if new_finished > 0:
                pbar.update(new_finished)
            else:
                pbar.refresh()
        time.sleep(1)

    if pbar:
        pbar.close()
    return results



from logging import Logger

from sklearn.base import ClassifierMixin
from sklearn import preprocessing, impute
import polars as pl

from xirescore.feature_extracting import get_features


def get_transformers(df: pl.DataFrame, options: dict):
    """
    Normalize the features and drop NaN-values if necessary.
    """
    filter_nan = options['rescoring']['imputer'] is None
    features = get_features(df, options, filter_nan)
    df_features = df.select(features)

    Scaler: ClassifierMixin.__class__ = getattr(preprocessing, options['rescoring']['scaler'])
    scaler_options = options['rescoring']['scaler_params']

    imputer = None
    if options['rescoring']['imputer'] is not None:
        # Impute missing feature values
        imputer_class = getattr(impute, options['rescoring']['imputer'])
        imputer = imputer_class(
            **options['rescoring']['imputer_kwargs']
        )
        imputer.fit(
            df_features
        )

    scaler = Scaler(**scaler_options)
    scaler.fit(df_features)

    return imputer, scaler, features



import argparse
import yaml
import ast
import logging
import os
import sys

import logging_loki

from xirescore.XiRescore import XiRescore
import xirescore

logger = logging.getLogger(__name__)

def main():
    # Create argument parser
    parser = argparse.ArgumentParser(description='Rescoring crosslinked-peptide identifications.')

    # Define CLI arguments
    parser.add_argument('-i', action='store', dest='input_path', help='input path',
                        type=str, required=False)
    parser.add_argument('-o', action='store', dest='output_path', help='output path',
                        type=str, required=False)
    parser.add_argument('-c', action='store', dest='config_file', help='config file',
                        type=str, required=False)
    parser.add_argument('-C', action='store', dest='config_string', help='config test',
                        type=str, required=False)
    parser.add_argument('--loki', action='store', dest='loki', help='Loki server address',
                        type=str, required=False)
    parser.add_argument('--loki-job-id', action='store', dest='loki_job_id', help='Loki job ID',
                        default="", type=str, required=False)
    parser.add_argument('--version', action='store_true', dest='print_version', help='print version')

    # Parse arguments
    args = parser.parse_args()

    if args.print_version:
        print(xirescore.__version__)
        os._exit(os.EX_OK)

    if args.input_path is None:
        print('xirescore: error: the following arguments are required: -i', file=sys.stderr)
        parser.print_help()
        os._exit(os.EX_USAGE)
    if args.output_path is None:
        print('xirescore: error: the following arguments are required: -o', file=sys.stderr)
        parser.print_help()
        os._exit(os.EX_USAGE)

    # Load config
    if args.config_file is not None:
        with open(args.config_file, 'r') as file:
            options = yaml.safe_load(file)
    elif args.config_string is not None:
        options = ast.literal_eval(args.config_string)
    else:
        options = dict()

    # Configure Loki logger
    if args.loki is None:
        logging.basicConfig()
        logger.setLevel(logging.DEBUG)
    else:
        handler = logging_loki.LokiHandler(
            url=args.loki,
            tags={
                "application": "xirescore",
                "job_id": args.loki_job_id
            },
            version="1",
        )
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    # Configure stdout logger
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # Initialize XiRescore
    rescorer = XiRescore(
        input_path=args.input_path,
        output_path=args.output_path,
        options=options,
    )
    rescorer.run()
    logger.info("Done.")


if __name__ == "__main__":
    main()



from collections.abc import Iterable
import uuid

import numpy as np


def serialize_columns(df):
    nan_filter = None
    for c in df.columns:
        if nan_filter is None:
            nan_filter = ~df[c].isna()
        else:
            nan_filter &= ~df[c].isna()

    # Try to find row without NaNs
    type_row = df[nan_filter]

    if len(type_row) > 0:
        # If some found take the first one
        type_row = type_row.iloc[:1]
    else:
        # If none found take the first row od the DF
        type_row = df.iloc[:1]

    for c in df.columns:
        col_row = type_row
        # Check if type_row has a valid value for column
        if col_row.iloc[0][c] is None:
            # If not search for a row with a non-NaN value
            col_row = df[~df[c].isna()]
            if len(col_row) == 0:
                # Only NaN values for column
                continue
        # Take the non-NaN value found
        col_row = col_row.iloc[0]
        if type(col_row[c]) is uuid.UUID:
            df[df[c].isna()] = ''
            df[c] = df[c].astype(str, errors='ignore')
        elif type(col_row[c]) is str:
            pass
        elif isinstance(col_row[c], Iterable):
            df[df[c].isna()] = ''
            df[c] = df[c].apply(
                lambda x: None if x is None else ';'.join(np.array(x).astype(str))
            )
        elif np.issubdtype(type(col_row[c]), np.generic):
            pass
        else:
            df[df[c].isna()] = ''
            df[c] = df[c].astype(str, errors='ignore')
    df.columns = df.columns.map(str)
    return df



import random
import re
import logging

import networkx
import pandas as pd
import numpy as np
import networkx as nx


logger = logging.getLogger(__name__)

class NoOverlapKFold:
    def __init__(self, n_splits: int = 5, shuffle: bool = False, random_state: int = 42,
                 pep1_id_col: str = "base_sequence_p1", pep2_id_col: str = "base_sequence_p2",
                 target_col='isTT'):
        """
        Constructor for NoOverlapKFold.

        Parameters:
        - n_splits (int, optional): Number of splits. Default is 5.
        - shuffle (bool, optional): Whether to shuffle the data before splitting. Default is False.
        - random_state (int or RandomState, optional): Seed for the random number generator. Default is None.
        - pep1_id_col (str, optional): Column name for PepSeq1. Default is "base_sequence_p1".
        - pep2_id_col (str, optional): Column name for PepSeq2. Default is "base_sequence_p2".
        """
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state
        self.pep1_id_col = pep1_id_col
        self.pep2_id_col = pep2_id_col
        self.target_col = target_col

    def splits_by_peptides(self,
                           df: pd.DataFrame,
                           labels: pd.DataFrame = None,
                           pepseqs: pd.DataFrame = None,
                           n_retries = 10):
        """
        Splits a DataFrame into multiple subsets based on peptide sequences for cross-validation.

        This function creates a graph-based network of peptide sequences and splits the DataFrame
        into a specified number of slices for cross-validation. It considers sequences that are
        only different by modifications and ensures that the same peptide sequences are grouped
        together.

        Parameters
        ----------
        df : pd.DataFrame
            The input DataFrame containing the peptide data.
        labels : pd.DataFrame, optional
            Labeling of the samples. If not provided, the 'df' DataFrame is used.
        pepseqs : pd.DataFrame, optional
            An optional DataFrame containing specific peptide sequences to be used for creating the
            graph. If not provided, the 'df' DataFrame is used.

        Returns
        -------
        List[Tuple[np.ndarray, np.ndarray]]
            A list of tuples, each containing two arrays: the indices of the training set and the
            indices of the test set for each split.

        Notes
        -----
        The function performs the following steps:
        1. Creates a graph for the peptide sequence network.
        2. Groups peptide sequences by disregarding modifications.
        3. Constructs a graph from the grouped sequences.
        4. Calculates the maximum slice size and splits the graph into components.
        5. Maps the grouped sequences back to peptides and recombines them into slices.
        6. Adds slice columns to the original dataset.
        7. Ensures no overlap between train and test sets in each split.

        The method assumes the existence of the following instance attributes:
        - `pep1_id_col`: The column name for the first peptide identifier.
        - `pep2_id_col`: The column name for the second peptide identifier.
        - `n_splits`: The number of splits to be created.
        - `random_state`: A random seed for reproducibility.
        - `logger`: A logger for debugging and information messages.

        Examples
        --------
        >>> splits = obj.splits_by_peptides(df)
        >>> for train_index, test_index in splits:
        >>>     print("TRAIN:", train_index, "TEST:", test_index)
        """
        random.seed(self.random_state)
        df_start = df.copy()
        if labels is None:
            labels = df[[self.target_col]]
        labels_start = labels.copy()
        if pepseqs is None:
            pepseqs = df[[
                self.pep1_id_col,
                self.pep2_id_col,
            ]]
        pepseqs_start = pepseqs.copy()
        for try_run in range(n_retries):
            df = df_start.copy()
            labels = labels_start.copy()
            pepseqs = pepseqs_start.copy()
            seed = random.randint(0, 2**32)

            # Create graph for CSM network
            edges = None
            if pepseqs is not None:
                edges = pepseqs[[self.pep1_id_col, self.pep2_id_col]]
            else:
                edges = df[[self.pep1_id_col, self.pep2_id_col]]

            # Pepseq grouping
            # In the beginning this will just be a enumerated list of all peptides
            pepseqs_grouping = pd.concat(
                [
                    edges[[self.pep1_id_col]].rename({self.pep1_id_col: 'pepseq'}, axis=1),
                    edges[[self.pep2_id_col]].rename({self.pep2_id_col: 'pepseq'}, axis=1),
                ]
            ).drop_duplicates()

            # Create column for converted sequences
            # This was initially intended as the unmodified peptide sequence. Even though we now use
            # the already unmodified base_sequence column, we keep it in here for compatibility.
            pepseqs_grouping['alt_pepseq'] = pepseqs_grouping['pepseq']

            # Initialize groups by assigning each peptide its own group
            pepseqs_grouping['group'] = pepseqs_grouping.reset_index().index

            # Merge peptide sequences that only differ by modifications
            logger.debug("Disregard modifications")
            pepseqs_grouping = self.regroup_unmodified(pepseqs_grouping)

            n_pepseqs = len(pepseqs_grouping['pepseq'].drop_duplicates())
            n_groups = len(pepseqs_grouping['group'].drop_duplicates())
            logger.debug(f"Grouping factor: {n_groups} groups / {n_pepseqs} seqs = {n_groups/n_pepseqs:.2f}")

            # Apply grouping
            logger.debug(f"Apply grouping")
            edges_grouped = edges.merge(
                pepseqs_grouping.rename({'group': 'source'}, axis=1),
                left_on=self.pep1_id_col,
                right_on="pepseq",
                validate='many_to_one',
            ).merge(
                pepseqs_grouping.rename({'group': 'target'}, axis=1),
                left_on=self.pep2_id_col,
                right_on="pepseq",
                validate='many_to_one',
            ).drop_duplicates()

            logger.debug(f"Construct graph")
            peptide_graph: networkx.Graph = nx.from_pandas_edgelist(edges_grouped)

            # Calculate maximum slice size
            slice_size = int((peptide_graph.number_of_nodes() / self.n_splits))
            logger.debug(f"Maximum slice size: {slice_size}")

            # Split into components and communities
            commties = self.recursive_async_fluidc(peptide_graph, max_size=slice_size, seed=seed)
            logger.debug("Clustering done.")

            # Map group commties back to peptides
            pepseq_commties = self.group_to_pepseqs(commties, pepseqs_grouping)

            # Recombine into slices
            slices = self.communities_slicing(pepseq_commties, self.n_splits)
            logger.debug(f"Slice sizes: {[len(s) for s in slices]}")

            # Convert to pandas dataframe
            slicing_parts = []
            for i, s in enumerate(slices):
                slicing_parts.append(pd.DataFrame(
                    [[peptide, i] for peptide in s],
                    columns=['kfold_peptide', 'slice']
                ))
            slicing = pd.concat(slicing_parts)

            # Add slice columns for both peptides to dataset
            edges = pd.merge(
                edges,
                slicing,
                left_on=self.pep1_id_col,
                right_on='kfold_peptide',
                how='left',
                validate='many_to_one',
            ).set_axis(edges.index).rename({'slice': 'slice1'}, axis=1)

            edges = pd.merge(
                edges,
                slicing,
                left_on=self.pep2_id_col,
                right_on='kfold_peptide',
                how='left',
                validate='many_to_one',
            ).set_axis(edges.index).rename({'slice': 'slice2'}, axis=1)

            # On slice spanning CSMs choose one pseudo randomly
            edges['slice1_hash'] = (edges[self.pep1_id_col] + str(self.random_state)).apply(hash)
            edges['slice2_hash'] = (edges[self.pep2_id_col] + str(self.random_state)).apply(hash)
            edges['decision_hash'] = (edges['slice1_hash']+edges['slice2_hash']).apply(hash)
            edges['slice1_dist'] = np.absolute(edges['slice1_hash'] - edges['decision_hash'])
            edges['slice2_dist'] = np.absolute(edges['slice2_hash'] - edges['decision_hash'])
            edges['slice'] = np.where(
                edges['slice1_dist'] < edges['slice2_dist'],
                edges['slice1'],
                edges['slice2']
            )

            df = df.merge(
                edges[['slice']],
                left_index=True,
                right_index=True,
                how='inner',
                validate='one_to_one',
            )
            splits = [
                (
                    df[df['slice'] != i].index,
                    df[df['slice'] == i].index
                )
                for i in range(self.n_splits)
            ]
            splits_clean = [
                None for _ in splits
            ]
            for i, s in enumerate(splits):
                old_size = len(splits[i][0])
                splits_clean[i] = self.cleanup_split(
                    s,
                    pepseqs,
                    seq1col=self.pep1_id_col,
                    seq2col=self.pep2_id_col,
                )
                new_size = len(splits_clean[i][0])
                logger.info(
                    f"Split {i} train set clean-up: {new_size}/{old_size}="
                    f"{(100 * new_size / old_size): .2f}% remaining"
                )

            logger.debug("===DEBUG===")
            logger.info("Sanity checks")
            testcum = set()
            labels_ok = True
            for i_split, (train_s, test_s) in enumerate(splits_clean):
                logger.debug(f"Train/Test: {len(train_s):,.0f}/{len(test_s):,.0f}")
                if len(np.intersect1d(train_s, test_s)) > 0:
                    logger.fatal("FATAL! Train and test overlapping")
                    return None
                train_label_counts = labels.loc[train_s, self.target_col].value_counts()
                test_label_counts = labels.loc[test_s, self.target_col].value_counts()
                logger.debug(f"Slice {i_split} train: {train_label_counts.to_dict()}")
                logger.debug(f"Slice {i_split} test: {test_label_counts.to_dict()}")
                if len(train_label_counts) < 2:
                    logger.warning(
                        f"Try {try_run+1} of {n_retries} failed: "
                        f"Only one training label in slice {i_split}."
                    )
                    labels_ok = False
                    break
                if len(test_label_counts) < 2:
                    logger.warning(
                        f"Try {try_run+1} of {n_retries} failed: "
                        f"Only one test label in slice {i_split}."
                    )
                    labels_ok = False
                    break
                testcum.update(list(test_s))

            if not labels_ok:
                continue

            if len(df) != len(testcum):
                logger.error(f"FATAL! Not all training data tested {len(testcum):,.0f} of {len(df):,.0f}.")
                return None

            return splits_clean
        logger.critical('Fatal: Could not find a working splitting.')

    def cleanup_split(self, fold, pepseqs, seq1col="sequence_p1", seq2col="sequence_p2"):
        train_index, test_index = fold
        unique_values_test = pd.concat([
            pepseqs.loc[test_index][seq1col],
            pepseqs.loc[test_index][seq2col]
        ]).drop_duplicates()
        conflict_idx = pepseqs.loc[train_index].loc[
            np.isin(pepseqs.loc[train_index][seq1col], unique_values_test) |
            np.isin(pepseqs.loc[train_index][seq2col], unique_values_test)
            ].index
        filtered_train_index = train_index.difference(conflict_idx)
        return filtered_train_index, test_index

    def _recursive_async_fluidc_comp(self, comp_g: nx.Graph, seed, max_size=1000000, n_communities=2) -> list[set]:
        good_communities = []
        commties = [c for c in nx.community.asyn_fluidc(comp_g, n_communities, seed=seed)]
        commty_counts = [len(c) for c in commties]
        logger.debug(f"Community sizes: {commty_counts} = {sum(commty_counts)}")
        for comm in commties:
            if len(comm) <= max_size:
                good_communities += [comm]
            else:
                comm_g = nx.subgraph(comp_g, comm)
                good_communities += self.recursive_async_fluidc(comm_g, max_size=max_size, seed=seed)
        return good_communities

    def recursive_async_fluidc(self, g: nx.Graph, seed, max_size=1000000, n_communities=2) -> list[set]:
        good_communities = []
        comps = [c for c in nx.connected_components(g)]
        logger.debug(f"Number of components: {len(comps):,.0f}")
        for comp in comps:
            if len(comp) <= max_size:
                good_communities += [comp]
            else:
                comp_g = nx.subgraph(g, comp)
                good_communities += self._recursive_async_fluidc_comp(
                    comp_g,
                    max_size=max_size,
                    n_communities=n_communities,
                    seed=seed
                )
        return good_communities

    def communities_slicing(self, commties, n_slices=2) -> list[set]:
        # Sort communities by size (ascending)
        commty_sizes = np.array([len(c) for c in commties])
        commties = [commties[i] for i in commty_sizes.argsort()]

        slices = [set() for _ in range(n_slices)]

        while len(commties) > 0:
            # Largest community
            commty = commties.pop()
            # Sort slices by size (ascending)
            slice_sizes = np.array([len(s) for s in slices])
            slices = [slices[i] for i in slice_sizes.argsort()]
            # Add to smallest slice
            slices[0].update(commty)

        return slices

    def regroup_unmodified(self, pepseq_grouping: pd.DataFrame) -> pd.DataFrame:
        pepseq_grouping['alt_pepseq'] = pepseq_grouping['alt_pepseq'].apply(
            lambda x: re.sub(r"[^A-Z]", "", x)
        )

        regrouping: pd.DataFrame = pepseq_grouping.groupby('alt_pepseq').min()
        regrouping.rename({'group': 'new_group'}, axis=1, inplace=True)

        pepseq_grouping = pepseq_grouping.merge(
            regrouping[['new_group']],
            left_on='alt_pepseq',
            right_index=True,
            validate='many_to_one',
        )

        pepseq_grouping['group'] = pepseq_grouping['new_group']

        return pepseq_grouping[['pepseq', 'alt_pepseq', 'group']]

    def group_to_pepseqs(self, commties: list[set], pepseqs_grouping: pd.DataFrame):
        pepseq_commties = []
        for c in commties:
            commty_filter = pepseqs_grouping['group'].isin(c)
            pepseqs = pepseqs_grouping.loc[commty_filter, 'pepseq']
            pepseq_commties += [set(pepseqs)]
        return pepseq_commties



from typing import Callable
import importlib
import logging

import polars as pl
from sklearn.base import ClassifierMixin
import sklearn
from sklearn.metrics import accuracy_score, balanced_accuracy_score
import multiprocess as mp

from xirescore import async_result_resolving
from xirescore.NoOverlapKFold import NoOverlapKFold


logger = logging.getLogger(__name__)

def train(train_df: pl.DataFrame, cols_features, clf_params, options, splits=None):
    # Get peptide sequence columns
    cols_pepseq = [
        options['input']['columns']['base_sequence_p1'],
        options['input']['columns']['base_sequence_p2'],
    ]

    # Get column for target labeling
    col_label = options['input']['columns']['is_tt']

    # Get DataFrames for peptide sequences, features and labels
    pepseq_df = train_df.select(cols_pepseq)
    features_df = train_df.select(cols_features)
    labels_df = train_df[col_label]

    # Import model
    model_class = options['rescoring']['model_class']
    model_name = options['rescoring']['model_name']
    model_module = importlib.import_module(f"sklearn.{model_class}")
    model: ClassifierMixin.__class__ = getattr(model_module, model_name)

    # Check for single core model
    is_mp_model = hasattr(model, 'n_jobs')

    # Get random seed and number of k-fold splits
    seed = options['rescoring']['random_seed']
    n_splits = options['rescoring']['n_splits']

    if splits is None:
        # Get non peptide overlapping k-fold splits
        kf = NoOverlapKFold(
            n_splits,
            pep1_id_col=options['input']['columns']['base_sequence_p1'],
            pep2_id_col=options['input']['columns']['base_sequence_p2'],
            target_col=col_label,
            random_state=seed,
            shuffle=True,
        )
        splits = kf.splits_by_peptides(
            df=train_df.to_pandas(),
            pepseqs=pepseq_df.to_pandas()
        )

    max_jobs = options['rescoring']['max_jobs']
    if max_jobs < 1:
        max_jobs = mp.cpu_count() - 1

    fold_clfs = []
    with mp.get_context("spawn").Pool(processes=max_jobs) as pool:
        for cvi, fold in enumerate(splits):
            # Only run multiprocessing for single core models
            if is_mp_model:
                fold_clf = [
                    lambda: train_fold(
                        features_df=features_df,
                        labels_df=labels_df,
                        fold=fold,
                        params=clf_params,
                        options=options
                    )
                ]
            else:
                fold_clf = pool.apply_async(
                    train_fold,
                    kwds=dict(
                        features_df=features_df,
                        labels_df=labels_df,
                        fold=fold,
                        params=clf_params,
                        options=options
                    )
                )
            # Add job to the
            fold_clfs.append(fold_clf)
        # Resolve (potentially) async results
        clfs = async_result_resolving.resolve(fold_clfs)
    return clfs, splits


def train_fold(features_df, labels_df, fold, params, options):
    # Import classifier model
    model_class = options['rescoring']['model_class']
    model_name = options['rescoring']['model_name']
    model_module = importlib.import_module(f"sklearn.{model_class}")
    model: ClassifierMixin.__class__ = getattr(model_module, model_name)

    # Import metric function
    metric_name = options['rescoring']['metric_name']
    metric: Callable = getattr(sklearn.metrics, metric_name)

    # Unpack fold
    train_idx, test_idx = fold

    # Get fold's train features and labels
    fold_train_features_df = features_df[train_idx.to_numpy()]
    fold_train_labels_df = labels_df[train_idx.to_numpy()]

    # Get fold's test features and labels
    fold_test_features_df = features_df[test_idx.to_numpy()]
    fold_test_labels_df = labels_df[test_idx.to_numpy()]

    # Train fold classifier
    clf = model(**params)
    clf.fit(
        fold_train_features_df,
        fold_train_labels_df
    )
    test_predictions = clf.predict(fold_test_features_df)

    # Evaluate fold model
    score = metric(
        fold_test_labels_df.to_numpy(),
        test_predictions,
        labels=[0, 1]
    )
    accuracy = accuracy_score(
        fold_test_labels_df.to_numpy(),
        test_predictions,
    )
    balanced_accuracy = balanced_accuracy_score(
        fold_test_labels_df.to_numpy(),
        test_predictions,
    )
    logger.info(f"Score: {score}")
    logger.info(f"Accuracy: {accuracy*100:.2f}%")
    logger.info(f"Balanced accuracy: {balanced_accuracy*100:.2f}%")

    return clf



import sys
import warnings
from typing import Callable
from functools import partial
import logging
import importlib
import multiprocess as mp

import numpy as np
import polars as pl
import sklearn
from sklearn.base import ClassifierMixin
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import accuracy_score, balanced_accuracy_score
from sklearn.model_selection import ParameterGrid

from xirescore import async_result_resolving
from xirescore.NoOverlapKFold import NoOverlapKFold

logger = logging.getLogger(__name__)

def get_hyperparameters(train_df: pl.DataFrame, cols_features, splits, options):
    # Get peptide sequence columns
    cols_pepseq = [
        options['input']['columns']['base_sequence_p1'],
        options['input']['columns']['base_sequence_p2'],
    ]

    # Get column for target labeling
    col_label = options['input']['columns']['is_tt']

    # Get DataFrames for peptide sequences, features and labels
    pepseq_df = train_df.select(cols_pepseq)
    features_df = train_df.select(cols_features)
    labels_df = train_df[col_label]

    # Import model
    model_class = options['rescoring']['model_class']
    model_name = options['rescoring']['model_name']
    model_module = importlib.import_module(f"sklearn.{model_class}")
    model: ClassifierMixin.__class__ = getattr(model_module, model_name)

    # Check for single core model
    is_mp_model = hasattr(model, 'n_jobs')

    # Get random seed and number of k-fold splits
    seed = options['rescoring']['random_seed']
    n_splits = options['rescoring']['n_splits']

    # Get the list of hyperparameter configurations
    hyperparam_grid = list(ParameterGrid(options['rescoring']['model_params']))

    if splits is None:
        # Get non peptide overlapping k-fold splits
        kf = NoOverlapKFold(
            n_splits,
            random_state=seed,
            shuffle=True,
            pep1_id_col=options['input']['columns']['base_sequence_p1'],
            pep2_id_col=options['input']['columns']['base_sequence_p2'],
            target_col=col_label,
        )
        splits = kf.splits_by_peptides(
            df=train_df.to_pandas(),
            pepseqs=pepseq_df.to_pandas()
        )

    max_jobs = options['rescoring']['max_jobs']
    if max_jobs < 1:
        max_jobs = mp.cpu_count()-1

    logger.info(f'Using {max_jobs} CPU cores')

    # Create partial function for parameter testing
    logger.debug(f"Metric name: {options['rescoring']['metric_name']}")
    param_try_job = partial(
        _try_parameters,
        features_df=features_df,
        labels_df=labels_df,
        splits=splits,
        options=options
    )

    with mp.get_context("spawn").Pool(processes=max_jobs) as pool:
        # Only run multiprocessing for single core models
        if is_mp_model:
            param_scores = [
                param_try_job(params=params)
                for params in hyperparam_grid
            ]
        else:
            param_scores = [
                pool.apply_async(
                    param_try_job,
                    kwds=dict(params=params)
                )
                for params in hyperparam_grid
            ]

        # Resolve (potentially) async results
        param_scores = async_result_resolving.resolve(param_scores)

    # Get the best parameters based on metric
    if options['rescoring']['minimize_metric']:
        best_params_i = np.argmin(param_scores)
    else:
        best_params_i = np.argmax(param_scores)

    best_params = hyperparam_grid[best_params_i]
    logger.info(f"Best hyperparameter: {best_params}")

    return best_params


def _try_parameters(features_df: pl.DataFrame,
                    labels_df: pl.DataFrame,
                    splits,
                    params,
                    options,):
    # Create child logger for parameter configuration
    logger.debug(f"Params: {params}")

    # Import classifier model
    model_class = options['rescoring']['model_class']
    model_name = options['rescoring']['model_name']
    model_module = importlib.import_module(f"sklearn.{model_class}")
    model: ClassifierMixin.__class__ = getattr(model_module, model_name)

    # Import metric function
    metric_name = options['rescoring']['metric_name']
    logger.debug(f"Metric name: {options['rescoring']['metric_name']}")
    metric: Callable = getattr(sklearn.metrics, metric_name)

    # Initialize result lists
    scores = []
    accuracies = []
    balanced_accuracies = []
    for cvi, fold in enumerate(splits):
        logger.debug(f"Fold {cvi}")

        # Unpack fold
        train_idx, test_idx = fold

        # Get fold's train features and labels
        fold_train_features_df = features_df[train_idx.to_numpy()]
        fold_train_labels_df = labels_df[train_idx.to_numpy()]

        # Get fold's test features and labels
        fold_test_features_df = features_df[test_idx.to_numpy()]
        fold_test_labels_df = labels_df[test_idx.to_numpy()]

        # Train fold classifier
        clf = model(**params)
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always", category=ConvergenceWarning)
            clf.fit(
                fold_train_features_df,
                fold_train_labels_df
            )
            if len(caught_warnings) > 0:
                print(f'Warnings for parameters: {params}', file=sys.stderr)
                for warn in caught_warnings:
                    print(warn.message, file=sys.stderr)
        test_predictions = clf.predict(fold_test_features_df)

        # Evaluate fold model
        score = metric(
            fold_test_labels_df.to_numpy(),
            test_predictions,
            labels=[0, 1]
        )
        accuracy = accuracy_score(
            fold_test_labels_df.to_numpy(),
            test_predictions,
        )
        balanced_accuracy = balanced_accuracy_score(
            fold_test_labels_df.to_numpy(),
            test_predictions,
        )

        # Store evaluation
        scores.append(score)
        accuracies.append(accuracy)
        balanced_accuracies.append(balanced_accuracy)

    # Log fold performances
    logger.debug(f"Scores: {scores}")
    logger.debug(f"Accuracies: {accuracies}")
    logger.debug(f"Balanced accuracies: {balanced_accuracies}")

    # Return mean score
    return np.mean(scores)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "frosch~=0.1.9",
    "pandas>=2.0.0",
    "numpy>=1.7.2",
    "PyYAML~=6.0",
    "tabulate~=0.9.0",
    "numpy-indexed~=0.3.7",
    "scikit-learn>=1.3.2",
    "joblib~=1.2.0",
    "xgboost~=1.7.5",
    "tqdm~=4.66.0",
    "hyperopt~=0.2.7",
    "lightgbm~=3.3.5",  # Check the pre-requirements https://pypi.org/project/lightgbm/
    "imblearn~=0.0",
    "py~=1.11.0",
    "networkx~=3.1",
    "multiprocess~=0.70.16",
    "threadpoolctl>=3.5.0",
    "deepmerge~=1.1.0",
    "SQLAlchemy~=2.0.30",
    "psycopg2>=2.9",
    "python-logging-loki",
    "fastparquet>=2022.11.0",
    "pyarrow",
    "scipy>=1.0.1",
    "setuptools~=71.1.0",
    "xiutilities~=1.2.3",
    "polars",
    "xifdr"
]

requirements_dev = [
    'pip>=19.2.3',
    'bump2version>=0.5.11',
    'wheel>=0.33.6',
    'watchdog>=0.9.0',
    'flake8>=3.7.8',
    'coverage>=4.5.4',
    'Sphinx>=1.8.5',
    'twine>=1.14.0',
    'setuptools_scm',
    'pytest>=6.2.4',
    'black>=21.7b0',
    'sphinx_rtd_theme',
]

requirements_test = [
    'pytest>=3',
    "pydocstyle~=6.3.0",
    "pytest-cov~=4.0.0",
    "pytest-flake8~=1.0.6",
    "pytest-pydocstyle~=2.3.2",
    "flake8==4.0.1",
    "pytest-xdist",
]

requirements_docs = [
    "sphinx",
    "sphinx_rtd_theme",
    "sphinx-favicon",
]

setup(
    author="Falk Boudewijn Schimweg",
    author_email='git@falk.schimweg.de',
    python_requires='>=3.9',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    description="A machine learning based approach to rescore crosslinked spectrum matches (CSMs).",
    entry_points={
        'console_scripts': [
            'xirescore=xirescore.cli:main',
        ],
    },
    install_requires=requirements,
    extras_require={
        'test': requirements_test,
        'docs': requirements_docs,
        'dev': requirements_dev + requirements_test + requirements_docs,
    },
    license="GNU Affero General Public License v3 or later (AGPLv3+)",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='xirescore',
    name='xirescore',
    packages=find_packages(include=['xirescore', 'xirescore.*'], exclude=["tests"]),
    test_suite='tests',
    tests_require=requirements_test,
    url='https://github.com/Rappsilber-Laboratory/xiRescore',
    zip_safe=False,
)



"""Main module."""
import copy
import logging
import random
from collections.abc import Collection
from math import ceil

import numpy as np
from networkx.algorithms.bipartite.cluster import modes
from sklearn.decomposition import PCA
import polars as pl
from deepmerge import Merger
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.inspection import permutation_importance

from xirescore import readers
from xirescore import rescoring
from xirescore import train_data_selecting
from xirescore import training
from xirescore import writers
from xirescore._default_options import default_options
from xirescore.column_generating import generate as generate_columns
from xirescore.feature_extracting import get_features
from xirescore.feature_scaling import get_transformers
from xirescore.hyperparameter_optimizing import get_hyperparameters

options_merger = Merger(
    # pass in a list of tuple, with the
    # strategies you are looking to apply
    # to each type.
    [
        (dict, ["merge"]),
        (list, ["override"]),
        (set, ["override"])
    ],
    # next, choose the fallback strategies,
    # applied to all other types:
    ["override"],
    # finally, choose the strategies in
    # the case where the types conflict:
    ["override"]
)

logger = logging.getLogger(__name__)


class XiRescore:
    def __init__(self,
                 input_path,
                 output_path=None,
                 options=dict()):
        """
        Initialize rescorer

        :param input_path: Path to input file/DB or an input DataFrame.
        :type input_path: str|DataFrame
        :param output_path: Path to the output file/DB or ``None`` if ``get_rescored_output()`` will be used.
        :type output_path: str, optional
        :param options: :ref:`options`
        :type options: dict, optional
        """
        # Apply override default options with user-supplied options
        self._options = copy.deepcopy(default_options)
        if 'model_params' in options.get('rescoring', dict()):
            # Discard default model_params if new ones are provided
            del self._options['rescoring']['model_params']
        self._options = options_merger.merge(
            self._options,
            options
        )

        # Set random seed
        seed = self._options['rescoring']['random_seed']
        self._true_random_seed = random.randint(0, 2**32-1)
        np.random.seed(seed)
        random.seed(seed)


        # Store input data path
        self._input = input_path

        if output_path is None:
            # Store output in new DataFrame if no path is given
            self._output = pl.DataFrame()
        else:
            self._output = output_path

        self._true_random_ctr = 0

        self.train_df: pl.DataFrame
        """
        Data used for k-fold cross-validation.
        """
        self.splits: Collection[tuple[list, list]] = []
        """
        K-fold splits of model training. Kept to not rescore training samples with models they have been trained on.
        """
        self.models: list[BaseEstimator] = []
        """
        Trained models from the f-fold cross-validation.
        """
        self.scaler: TransformerMixin = None
        """
        PCA for feature decorrelation
        """
        self.pca: PCA = None
        """
        Scaler for feature normalization.
        """
        self.train_features: list = []
        """
        Features extracted from training data.
        """
        self.imputer: TransformerMixin = None
        """
        Imputer for missing values.
        """

    def run(self) -> None:
        """
        Run training and rescoring of the input data and write to output
        """
        logger.info("Start full train and rescore run")
        self.train()
        self.rescore()

    def train(self,
              train_df: pl.DataFrame = None,
              splits: list[tuple[list, list]] = None):
        """
        Run training on input data or on the passed DataFrame if provided.

        :param train_df: Data to be used training instead of input data.
        :type train_df: DataFrame, optional
        :param splits: K-fold splits for manual ``train_df``.
        :type splits: Index, optional
        """
        logger.info('Start training')

        if train_df is None:
            self.train_df, self.imputer, self.scaler, self.train_features = train_data_selecting.select(
                self._input,
                self._options,
            )
        else:
            self.train_df = train_df
            self.train_df = generate_columns(
                self.train_df,
                options=self._options,
                do_fdr=True,
                do_self_between=True
            )
            self.imputer, self.scaler, self.train_features = get_transformers(train_df, self._options)

        if splits is not None:
            self.splits = splits

        # Scale features
        train_df_transformed = self.train_df.clone()
        if self.imputer is not None:
            train_df_transformed[self.train_features] = self.imputer.transform(
                train_df_transformed[self.train_features]
            )
        train_df_transformed[self.train_features] = self.scaler.transform(
            train_df_transformed[self.train_features]
        )
        train_df_transformed = train_df_transformed.fill_nan(0)

        passed_feaures=self.train_features
        if self._options['rescoring']['pca_n_components'] is not None:
            self.pca = PCA(n_components=self._options['rescoring']['pca_n_components'])
            pca_features = self.pca.fit_transform(train_df_transformed[self.train_features])
            passed_feaures = [f'_pca_feature_{i}' for i  in range(pca_features.shape[1])]
            logger.info(f"PCA returned {len(passed_feaures)} features")
            train_df_transformed[passed_feaures] = pca_features


        logger.info("Perform hyperparameter optimization")
        model_params = get_hyperparameters(
            train_df=train_df_transformed,
            cols_features=passed_feaures,
            splits=splits,
            options=self._options,
        )

        logger.info("Train models")
        self.models, self.splits = training.train(
            train_df=train_df_transformed,
            cols_features=passed_feaures,
            clf_params=model_params,
            splits=splits,
            options=self._options,
        )

    def get_rescoring_state(self) -> dict:
        """
        Get state of the current instance to use it later to recreate identical instance.

        :returns: Models and k-fold slices
        :rtype: dict
        """
        cols_csm = self._options['input']['columns']['csm_id']
        cols_spectra = self._options['input']['columns']['spectrum_id']
        train_cols = list(set(cols_csm+cols_spectra))

        return {
            'options': self._options,
            'pca': self.pca,
            'imputer': self.imputer,
            'scaler': self.scaler,
            'training_data': self.train_df.select(train_cols),
            'train_features': self.train_features,
            'splits': self.splits,
            'models': self.models,
        }

    def _true_random(self, min_val=0, max_val=2**32-1):
        state = random.getstate()
        random.seed(self._true_random_seed+self._true_random_ctr)
        self._true_random_ctr += 1
        val = random.randint(min_val, max_val)
        random.setstate(state)
        return val

    def rescore(self) -> None:
        """
        Run rescoring on input data.
        """
        logger.info('Start rescoring')
        cols_spectra = self._options['input']['columns']['spectrum_id']
        spectra_batch_size = self._options['rescoring']['spectra_batch_size']

        # Read spectra list
        spectra = readers.read_spectra_ids(
            self._input,
            cols_spectra,
        )

        # Sort spectra
        spectra.sort()

        # Calculate number of batches
        n_batches = ceil(len(spectra)/spectra_batch_size)
        logger.info(f'Rescore in {n_batches} batches')

        # Iterate over spectra batches
        df_rescored = pl.DataFrame()
        for i_batch in range(n_batches):
            # Define batch borders
            spectra_range = spectra[
                i_batch*spectra_batch_size:(i_batch+1)*spectra_batch_size
            ]
            spectra_from = spectra_range[0]
            spectra_to = spectra_range[-1]
            logger.info(f'Start rescoring spectra batch {i_batch+1}/{n_batches} with `{spectra_from}` to `{spectra_to}`')

            # Generate schema overrides
            float_cols = self._options['input']['columns']['features']
            float_cols.append(
                self._options['input']['columns']['score']
            )
            schema_overrides = {
                c: pl.Float64
                for c in float_cols
            }

            # Read batch
            df_batch = readers.read_spectra_range(
                input=self._input,
                spectra_from=spectra_from,
                spectra_to=spectra_to,
                spectra_cols=cols_spectra,
                sequence_p2_col=self._options['input']['columns']['base_sequence_p2'],
                only_pairs=True,
                schema_overrides=schema_overrides
            )
            logger.info(f'Batch contains {len(df_batch):,.0f} samples')
            logger.debug(f'Batch uses approx. {df_batch.estimated_size("mb"):,.2f}MB of RAM')

            # Rescore batch
            df_batch = self.rescore_df(df_batch)

            # Store collected matches
            logger.info('Write out batch')
            if type(self._output) is pl.DataFrame:
                df_rescored = pl.concat([
                    df_rescored,
                    df_batch
                ])
            else:
                writers.append_rescorings(
                    self._output,
                    df_batch
                )

        # Keep rescored matches when no output is defined
        if type(self._output) is pl.DataFrame:
            self._output = df_rescored

    def rescore_df(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        Rescore a DataFrame of CSMs.

        :param df: CSMs to be rescored
        :type df: DataFrame

        :return: Rescored CSMs
        :rtype: DataFrame
        """
        cols_spectra = self._options['input']['columns']['spectrum_id']
        col_rescore = self._options['output']['columns']['rescore']
        col_top_ranking = self._options['input']['columns']['top_ranking']
        max_jobs = self._options['rescoring']['max_jobs']
        apply_logit = self._options['rescoring']['logit_result']
        if self._options['input']['columns']['csm_id'] is None:
            col_csm = list(self.train_df.columns)
        else:
            col_csm = self._options['input']['columns']['csm_id']

        # Impute missing feature values
        df_features = df[self.train_features]
        if self.imputer is not None:
            df_features = self.imputer.transform(df_features)

        # Scale features
        df_scaled_features = self.scaler.transform(df_features)
        df_scaled_features = np.nan_to_num(df_scaled_features)
        passed_feaures = self.train_features
        if self.pca is not None:
            df_scaled_features = self.pca.transform(df_scaled_features)
            passed_feaures = [f'_pca_feature_{i}' for i in range(df_scaled_features.shape[1])]
        df_scaled_features = pl.DataFrame(
            df_scaled_features,
            schema=passed_feaures
        )

        # Rescore DF
        df_scores = rescoring.rescore(
            models=self.models,
            df=df_scaled_features,
            rescore_col=col_rescore,
            apply_logit=apply_logit,
            max_cpu=max_jobs
        )

        logger.info('Merge new scores into original data')

        # Rescore training data only with test fold classifier
        logger.info('Reconstruct training data slices')
        cols_merge = list(set(col_csm+cols_spectra))
        df_slice = self.train_df.select(cols_merge)
        df_slice = df_slice.with_columns(
            pl.lit(-1).alias(f'{col_rescore}_slice')
        )
        for i, (_, idx_test) in enumerate(self.splits):
            # TODO move the slice assignment into `training()`
            df_slice = df_slice.with_columns(
                pl.when(
                    pl.int_range(len(df_slice)).is_in(idx_test)
                ).then(pl.lit(i)).otherwise(
                    pl.col(f'{col_rescore}_slice')
                ).alias(f'{col_rescore}_slice')
            )

        logger.info('Add merge columns to scores DataFrame')
        df_scores = pl.concat([
            df,
            df_scores,
        ], how="horizontal")

        logger.info('Merge slice info into batch')
        df_scores = df_scores.join(
            df_slice,
            on=cols_merge,
            how='left',
            maintain_order='left'
        )
        df_scores = df_scores.with_columns(
            pl.col(f'{col_rescore}_slice').fill_null(-1)
        )

        logger.info('Pick the correct score')
        score_ser = df_scores.select(
            pl.col(f'^{col_rescore}_[0-9]+$')
        ).cast(pl.List(pl.Float64)).select(
            pl.concat_list(pl.col('*'))
        ).to_series()
        df_scores = df_scores.with_columns(
            scores_list=score_ser
        ).with_columns(
            pl.when(
                pl.col(f'{col_rescore}_slice') > -1
            ).then(
                pl.col('scores_list').list.get(
                    pl.col(f'{col_rescore}_slice')
                )
            ).otherwise(
                pl.col(col_rescore)
            ).alias(col_rescore)
        ).drop('scores_list')

        # Calculate top_ranking
        logger.info('Calculate top ranking scores')
        df_top_rank = df_scores.group_by(cols_spectra).agg(
            pl.max(f'{col_rescore}').alias(f'{col_rescore}_max'),
            pl.min(f'{col_rescore}').alias(f'{col_rescore}_min'),
            pl.col(col_rescore).neg().rank('dense').alias(f'{col_rescore}_rank'),
            pl.col(col_rescore),
        ).explode([
            col_rescore,
            f'{col_rescore}_rank'
        ]).unique()
        df_scores = df_scores.join(
            df_top_rank,
            on=list(cols_spectra)+[col_rescore],
            maintain_order='left'
        )
        df_scores = df_scores.with_columns(
            (
                pl.col(col_rescore) == pl.col(f'{col_rescore}_max')
            ).alias(f'{col_rescore}_{col_top_ranking}')
        )

        return df_scores

    def get_rescored_output(self) -> pl.DataFrame:
        """
        Get the rescoring results when no output was defined

        :returns: Rescoring results
        :rtype: DataFrame
        """
        if type(self._output) is pl.DataFrame:
            return self._output
        else:
            raise XiRescoreError('Not available for file output.')

    def get_feature_importance(self) -> np.array:
        """
        Get the feature importances.
        """
        if not self.models:
            raise XiRescoreError('Models have not been trained.')
        if self.train_df is None:
            raise XiRescoreError('No training data defined.')

        # Check how to calculate feature importance
        importances = []
        for model in self.models:
            if hasattr(model, 'coef_'):
                importances.append(self._get_feature_importance_linear(model))
            elif hasattr(model, 'feature_importances_'):
                importances.append(self._get_feature_importance_tree(model))
            else:
                importances.append(self._get_feature_importance_permutation(model))

        if self.pca is not None:
            loadings = self.pca.components_.T
            orig_importances = []
            for importances_i in importances:
                orig_importances.append(loadings @ importances_i)
            importances = np.array(orig_importances)
        else:
            importances = np.array(importances)

        return importances

    def _get_feature_importance_linear(self, model) -> np.array:
        coef = model.coef_
        if coef.ndim == 1:
            importance = coef
        elif coef.ndim == 2:
            importance = np.mean(np.abs(coef), axis=0)
        else:
            raise ValueError(f"Unexpected coef_ shape: {coef.shape}")
        return importance

    def _get_feature_importance_tree(self, model) -> np.array:
        return model.feature_importances_

    def _get_feature_importance_permutation(self, model) -> np.array:
        features = self.train_df.select(self.train_features)
        if self.pca is not None:
            features = self.pca.transform(features)
        features = self.scaler.transform(features)
        labels = self.train_df[self._options['input']['columns']['is_tt']]
        result = permutation_importance(
            model,
            features,
            labels,
            n_repeats=10,
            random_state=self._options['rescoring']['random_seed'],
            n_jobs=self._options['rescoring']['max_jobs']
        )
        return result.importances_mean


def _select_right_score(row, col_rescore):
    n_slice = int(row[f"{col_rescore}_slice"])
    return row[f'{col_rescore}_{n_slice}']


class XiRescoreError(Exception):
    """Custom exception for train data selection errors."""
    pass



"""
Writers for data outputs
"""
from pathlib import Path
import random
import os

import fastparquet
import pandas as pd
import polars as pl

from xirescore.readers import get_source_type
from xirescore.df_serializing import serialize_columns


def append_rescorings(output, df: pd.DataFrame):
    output_type = get_source_type(output)
    if output_type == 'csv':
        append_csv(output, df)
    if output_type == 'tsv':
        append_csv(output, df, sep='\t')
    if output_type == 'parquet':
        append_parquet(output, df)


def append_parquet(output, df: pd.DataFrame|pl.DataFrame, compression='GZIP'):
    if type(df) is pl.DataFrame:
        df = df.to_pandas()
    else:
        df = serialize_columns(df)
    fastparquet.write(
        output,
        data=df,
        compression=compression,
        write_index=False,
        append=Path(output).is_file(),
    )


def append_csv(output, df: pd.DataFrame|pl.DataFrame, sep=','):
    if type(df) is pl.DataFrame:
        df = df.to_pandas()
    else:
        df = serialize_columns(df)
    df.to_csv(
        output,
        mode='a',
        header=not os.path.isfile(output),
        sep=sep
    )
